@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2014 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2014-2015 -- leonerd@leonerd.org.uk
 
 package Net::Async::Matrix;
 
@@ -11,7 +11,7 @@ use warnings;
 use base qw( IO::Async::Notifier );
 IO::Async::Notifier->VERSION( '0.63' ); # adopt_future
 
-our $VERSION = '0.13';
+our $VERSION = '0.14';
 $VERSION = eval $VERSION;
 
 use Carp;
@@ -552,7 +552,7 @@ sub start_longpoll
    $self->stop_longpoll;
    $self->{longpoll_last_token} = $args{start} // "END";
 
-   $self->{longpoll_f} = repeat {
+   my $f = $self->{longpoll_f} = repeat {
       my $last_token = $self->{longpoll_last_token};
 
       my $uri = $self->_uri_for_path( "/events",
@@ -579,6 +579,13 @@ sub start_longpoll
          $self->{make_delay}->( 3 )
       });
    } while => sub { !shift->failure };
+
+   # Don't ->adopt_future this one as it makes it hard to grab to cancel it
+   # again, but apply the same on_fail => invoke_error logic
+   $f->on_fail( $self->_capture_weakself( sub {
+      my $self = shift;
+      $self->invoke_error( @_ );
+   }));
 }
 
 sub stop_longpoll
@@ -839,22 +846,10 @@ sub join_room
    my $self = shift;
    my ( $room_alias ) = @_;
 
-   my $f;
-   if( $room_alias =~ m/^#/ ) {
-      $f = $self->_do_POST_json( "/join/$room_alias", {} )->then( sub {
-         my ( $content ) = @_;
-         Future->done( $content->{room_id} );
-      });
-   }
-   elsif( $room_alias =~ m/^!/ ) {
-      # Internal room ID directly
-      $f = $self->_do_PUT_json( "/rooms/$room_alias/state/m.room.member/$self->{user_id}", {
-         membership => "join",
-      } )->then_done( $room_alias );
-   }
+   $self->_do_POST_json( "/join/$room_alias", {} )->then( sub {
+      my ( $content ) = @_;
+      my $room_id = $content->{room_id};
 
-   $f->then( sub {
-      my ( $room_id ) = @_;
       if( my $room = $self->{rooms_by_id}{$room_id} ) {
          return Future->done( $room );
       }
@@ -979,6 +974,16 @@ sub _handle_event_m_room
       pp( $event ) );
 }
 
+sub _handle_event_m_typing
+{
+   my $self = shift;
+   my $event = pop;
+
+   if( my $room = $self->{rooms_by_id}{$event->{room_id}} ) {
+      $room->_handle_event_m_typing( $event );
+   }
+}
+
 sub _handle_roomevent_member_forward
 {
    my $self = shift;
