@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2014 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2014-2015 -- leonerd@leonerd.org.uk
 
 package Net::Async::Matrix::Room;
 
@@ -11,12 +11,13 @@ use warnings;
 # Not really a Notifier but we like the ->maybe_invoke_event style
 use base qw( IO::Async::Notifier );
 
-our $VERSION = '0.13';
+our $VERSION = '0.14';
 $VERSION = eval $VERSION;
 
 use Carp;
 
 use Future;
+use Future::Utils qw( repeat );
 
 use List::Util qw( pairmap );
 use Struct::Dumb;
@@ -24,6 +25,8 @@ use Time::HiRes qw( time );
 
 struct Member => [qw( user displayname membership )];
 
+use constant TYPING_RESEND_SECONDS => 30;
+
 =head1 NAME
 
 C<Net::Async::Matrix::Room> - a single Matrix room
@@ -104,6 +107,17 @@ C<$member> object will already be in the new state. C<%changes> will be a
 key/value list of state fields names that were changed, and references to
 2-element ARRAYs containing the old and new values for this field.
 
+=head2 on_typing $member, $is_typing
+
+Invoked on receipt of a typing notification change, when the given member
+either starts or stops typing.
+
+=head2 on_members_typing @members
+
+Invoked on receipt of a typing notification change to give the full set of
+currently-typing members. This is invoked after the individual C<on_typing>
+events.
+
 =cut
 
 sub _init
@@ -122,6 +136,10 @@ sub _init
    # separately, rather than on the member objects themselves
    $self->{powerlevels} = {};
 
+   # Server gives us entire sets of typing user_ids all at once. We have to
+   # remember state
+   $self->{typing_members} = {};
+
    $self->{aliases_by_hs} = {};
 }
 
@@ -131,7 +149,8 @@ sub configure
    my %params = @_;
 
    foreach (qw( on_message on_back_message on_membership on_back_membership
-         on_presence on_synced_state on_state_changed on_back_state_changed )) {
+         on_presence on_synced_state on_state_changed on_back_state_changed
+         on_typing on_members_typing )) {
       $self->{$_} = delete $params{$_} if exists $params{$_};
    }
 
@@ -972,6 +991,63 @@ sub paginate_messages
    $self->adopt_future( $f );
 }
 
+=head2 $room->typing_start
+
+Sends a typing notification that the user is currently typing in this room.
+This notification will periodically be re-sent as required by the protocol
+until the C<typing_stop> method is called.
+
+=cut
+
+sub typing_start
+{
+   my $self = shift;
+
+   return if $self->{typing_timer};
+
+   my $user_id = $self->{matrix}->myself->user_id;
+
+   my $f = $self->{typing_timer} = repeat {
+      $self->_do_PUT_json( "/typing/$user_id", {
+         typing  => 1,
+         timeout => ( TYPING_RESEND_SECONDS + 5 ) * 1000, # msec
+      })->then( sub {
+         $self->{matrix}->{make_delay}->( TYPING_RESEND_SECONDS );
+      });
+   } while => sub { !shift->failure };
+
+   $f->on_fail( $self->_capture_weakself( sub {
+      my $self = shift;
+      $self->invoke_error( @_ );
+   }));
+}
+
+=head2 $room->typing_stop
+
+Sends a typing notification that the user is no longer typing in this room.
+This method also cancels the repeating re-send behaviour created by
+C<typing_start>.
+
+=cut
+
+sub typing_stop
+{
+   my $self = shift;
+
+   return unless my $f = $self->{typing_timer};
+
+   $f->cancel;
+   undef $self->{typing_timer};
+
+   my $user_id = $self->{matrix}->myself->user_id;
+
+   $self->adopt_future(
+      $self->_do_PUT_json( "/typing/$user_id", {
+         typing => 0,
+      })
+   );
+}
+
 sub _handle_roomevent_create_forward
 {
    my $self = shift;
@@ -1025,6 +1101,33 @@ sub _handle_event_m_presence
    $self->maybe_invoke_event( on_presence => $member, %changes );
 }
 
+sub _handle_event_m_typing
+{
+   my $self = shift;
+   my ( $event ) = @_;
+
+   my $typing = $self->{typing_members};
+   my %not_typing = %$typing;
+
+   foreach my $user_id ( @{ $event->{content}{user_ids} } ) {
+      delete $not_typing{$user_id};
+      next if $typing->{$user_id};
+
+      $typing->{$user_id}++;
+      my $member = $self->{members_by_userid}{$user_id} or next;
+      $self->maybe_invoke_event( on_typing => $member, 1 );
+   }
+
+   foreach my $user_id ( keys %not_typing ) {
+      my $member = $self->{members_by_userid}{$user_id} or next;
+      $self->maybe_invoke_event( on_typing => $member, 0 );
+      delete $typing->{$user_id};
+   }
+
+   my @members = map { $self->{members_by_userid}{$_} } keys %$typing;
+   $self->maybe_invoke_event( on_members_typing => grep { defined } @members );
+}
+
 =head1 MEMBERSHIP STRUCTURES
 
 Parameters documented as C<$member> receive a membership struct, which
